/**
 * \file input/xtpxpipsaxhandler.cpp
 * \date 17/11/2017
 * \author Olivier Langella
 * \brief parse new xpip files from xtpcpp
 */


/*******************************************************************************
* Copyright (c) 2017 Olivier Langella <Olivier.Langella@u-psud.fr>.
*
* This file is part of XTPcpp.
*
*     XTPcpp is free software: you can redistribute it and/or modify
*     it under the terms of the GNU General Public License as published by
*     the Free Software Foundation, either version 3 of the License, or
*     (at your option) any later version.
*
*     XTPcpp is distributed in the hope that it will be useful,
*     but WITHOUT ANY WARRANTY; without even the implied warranty of
*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*     GNU General Public License for more details.
*
*     You should have received a copy of the GNU General Public License
*     along with XTPcpp.  If not, see <http://www.gnu.org/licenses/>.
*
* Contributors:
*     Olivier Langella <Olivier.Langella@u-psud.fr> - initial API and implementation
******************************************************************************/

#include "xtpxpipsaxhandler.h"

#include <pappsomspp/exception/exceptionnotfound.h>
#include <cmath>
#include "../core/peptideevidence.h"
#include "../utils/peptidestore.h"
#include "../utils/proteinstore.h"

XtpXpipSaxHandler::XtpXpipSaxHandler(WorkMonitorInterface * p_monitor, Project * p_project):_p_project(p_project)
{
    qDebug() << "XtpXpipSaxHandler::XtpXpipSaxHandler begin" ;
    _p_monitor = p_monitor;
    qDebug() << "XtpXpipSaxHandler::XtpXpipSaxHandler end" ;
}

XtpXpipSaxHandler::~XtpXpipSaxHandler()
{

}


bool XtpXpipSaxHandler::isXtpXpip() const {
    return _is_xtpcpp_xpip;
}
bool XtpXpipSaxHandler::startElement(const QString & namespaceURI, const QString & localName,
                                     const QString & qName, const QXmlAttributes & attributes) {
    qDebug()<< "XtpXpipSaxHandler::startElement begin" << namespaceURI << " " << localName << " " << qName ;
    _tag_stack.push_back(qName);
    bool is_ok = true;

    try {
        //startElement_group
        if (_tag_stack.size() == 1) {
            _is_xtpcpp_xpip = true;
            if (qName != "xpip") {
                _is_xtpcpp_xpip = false;
                throw pappso::ExceptionNotFound(QObject::tr("ERROR this file is not an xpip file %1").arg(qName));
            }
        } else if (qName == "protein_match") {
            is_ok = startElement_protein_match(attributes);
        } else if (qName == "peptide_match") {
            is_ok = startElement_peptide_match(attributes);
        } else if (qName == "protein") {
            is_ok = startElement_protein(attributes);
        } else if (qName == "identification_source") {
            is_ok = startElement_identification_source(attributes);
        }
        //<sample value="P6_08_10"/>
        else if (qName == "sample") {
            is_ok = startElement_msrun(attributes);
        } else if (qName == "peptide") {
            is_ok = startElement_peptide(attributes);
        } else if (qName == "peptide_evidence") {
            is_ok = startElement_peptide_evidence(attributes);
        } else if (qName == "modification") {
            is_ok = startElement_modification(attributes);
        } else if (qName == "mod") {
            is_ok = startElement_mod(attributes);
        }  else if (qName == "filter_params") {
            is_ok = startElement_filter_params(attributes);
        }  else if (qName == "description") {
            is_ok = startElement_description(attributes);
        } else if (qName == "fasta_file") {
            is_ok = startElement_fasta_file(attributes);
        }

        _current_text.clear();
    }
    catch (pappso::PappsoException exception_pappso) {
        _errorStr = QObject::tr("ERROR in XtpXpipSaxHandler::startElement tag %1, PAPPSO exception:\n%2").arg(qName).arg(exception_pappso.qwhat());
        return false;
    }
    catch (std::exception exception_std) {
        _errorStr = QObject::tr("ERROR in XtpXpipSaxHandler::startElement tag %1, std exception:\n%2").arg(qName).arg(exception_std.what());
        return false;
    }
    return is_ok;
}

bool XtpXpipSaxHandler::endElement(const QString & namespaceURI, const QString & localName,
                                   const QString & qName) {

    qDebug()<< "XtpXpipSaxHandler::endElement begin" << namespaceURI << " " << localName << " " << qName ;
    bool is_ok = true;
    // endElement_peptide_list
    try {
        if (qName == "protein")
        {
            is_ok = endElement_protein();
        }
        else if (qName == "peptide") {
            is_ok = endElement_peptide();
        }
        else if (qName == "sequence") {
            is_ok = endElement_sequence();
        }
        else if (qName == "protein_match") {
            is_ok = endElement_protein_match();
        }

        // end of detection_moulon
        // else if ((_tag_stack.size() > 1) &&
        //         (_tag_stack[_tag_stack.size() - 2] == "detection_moulon"))
    }
    catch (pappso::PappsoException exception_pappso) {
        _errorStr = QObject::tr("ERROR in XtpXpipSaxHandler::endElement tag %1, PAPPSO exception:\n%2").arg(qName).arg(exception_pappso.qwhat());
        return false;
    }
    catch (std::exception exception_std) {
        _errorStr = QObject::tr("ERROR in XtpXpipSaxHandler::endElement tag %1, std exception:\n%2").arg(qName).arg(exception_std.what());
        return false;
    }

    _current_text.clear();
    _tag_stack.pop_back();

    return is_ok;
}


bool XtpXpipSaxHandler::startElement_filter_params(QXmlAttributes attributes) {

//<filter_params pep_evalue="0.01" prot_evalue="-2.0" pep_number="1" filter_to_all="false" database_filter="contaminants_standarts.fasta"/>
    qDebug() << "startElement_filter_params ";
    _automatic_filter_parameters.setFilterPeptideEvalue( attributes.value("pep_evalue").simplified().toDouble());
    _automatic_filter_parameters.setFilterProteinEvalue( std::pow ((double) 10.0,attributes.value("prot_evalue").simplified().toDouble()));
    _automatic_filter_parameters.setFilterMinimumPeptidePerMatch( attributes.value("pep_number").simplified().toUInt());
    _automatic_filter_parameters.setFilterCrossSamplePeptideNumber(false);
    if (attributes.value("filter_to_all").simplified() == "true") {
        _automatic_filter_parameters.setFilterCrossSamplePeptideNumber(true);
    }
    _p_project->getFastaFileStore().getInstance(FastaFile(attributes.value("database_filter")));
    qDebug() << "startElement_filter_params end" ;
    return true;
}

bool XtpXpipSaxHandler::startElement_description(QXmlAttributes attributes) {

//    <description version="0.1.23" grouping="peptidemass" combine="true" date="2017-11-17T16:02:47"/>

    qDebug() << "startElement_description ";
    _p_project->setCombineMode(false);
    if (attributes.value("combine") == "true") {
        _p_project->setCombineMode(true);
    }
    //_total_protein_match = attributes.value("match_number").toUInt();

    //_p_monitor->setProgressMaximumValue(_total_protein_match);
    qDebug() << "startElement_description end" ;
    return true;
}

bool XtpXpipSaxHandler::startElement_modification(QXmlAttributes attributes) {

    /*
    <modification id="moda1" mod="MOD:00719"/>
    */
    qDebug() << "startElement_modifs_mass ";

    pappso::AaModificationP mod = pappso::AaModification::getInstance(attributes.value("mod").simplified());
    _map_modifs.insert(std::pair<QString, pappso::AaModificationP>(attributes.value("id").simplified(), mod));

    qDebug() << "startElement_modifs_mass end" ;
    return true;
}

bool XtpXpipSaxHandler::startElement_identification_source(QXmlAttributes attributes) {
    //<identification_source id="identa0" msrun_id="sampa0" path="/gorgone/pappso/formation/Janvier2014/TD/xml_tandem/20120906_balliau_extract_1_A01_urnb-1.xml" engine="1" version=""/>

    qDebug() << "startElement_identification_source ";

    IdentificationEngine engine = static_cast<IdentificationEngine>(attributes.value("engine").toUInt());
    IdentificationDataSourceSp sp_ident_source =  _p_project->getIdentificationDataSourceStore().getInstance(attributes.value("path").simplified(), engine);
    sp_ident_source.get()->setXmlId(attributes.value("id").simplified());
    _map_ident_sources.insert(std::pair<QString, IdentificationDataSourceSp>(sp_ident_source.get()->getXmlId(), sp_ident_source));
    
    sp_ident_source->setMsRunSp(_map_msruns.at(attributes.value("msrun_id").simplified()));
    sp_ident_source->setIdentificationEngineVersion(attributes.value("version"));
    //_current_identification_group_p->addMsRunSp(ms_run);
    qDebug() << "startElement_identification_source end" ;
    return true;
}
bool XtpXpipSaxHandler::startElement_msrun(QXmlAttributes attributes) {
//<msrun id="sampa0" name="20120906_balliau_extract_1_A01_urnb-1" format="0" path="20120906_balliau_extract_1_A01_urnb-1"/>

    qDebug() << "startElement_msrun ";

    MsRunSp ms_run =  _p_project->getMsRunStore().getInstance(attributes.value("path").simplified());
    ms_run.get()->setFilename(attributes.value("path").simplified());
    ms_run.get()->setSampleName(attributes.value("name").simplified());
    ms_run.get()->setXmlId(attributes.value("id").simplified());
    _map_msruns.insert(std::pair<QString, MsRunSp>(ms_run.get()->getXmlId(), ms_run));
    //_current_identification_group_p->addMsRunSp(ms_run);
    qDebug() << "startElement_msrun end" ;
    return true;
}

bool XtpXpipSaxHandler::startElement_peptide_evidence(QXmlAttributes attributes) {

    qDebug() << "startElement_peptide_evidence ";
    /*
     * <peptide_evidence id="peb76" source_id="identa0" peptide_id="pd7688" scan="11743" rt="2170" evalue="8.1e-10" exp_mass="1655.755648" charge="2" checked="true">
                    <param key="0" value="48.5"/>
                </peptide_evidence>

              */
    IdentificationDataSourceSp sp_ident_source = _map_ident_sources.at(attributes.value("source_id").simplified());
    unsigned int scan = attributes.value("scan").toUInt();
    _p_peptide_evidence = new PeptideEvidence(sp_ident_source.get()->getMsRunSp().get(), scan);
    _p_peptide_evidence->setIdentificationDataSource(sp_ident_source.get());
    
    _p_peptide_evidence->setChecked(false);
    if (attributes.value("checked").simplified().toLower() == "true") {
        _p_peptide_evidence->setChecked(true);
    }
    _p_peptide_evidence->setCharge(attributes.value("charge").toUInt());
    _p_peptide_evidence->setRetentionTime(attributes.value("rt").toDouble());
    _p_peptide_evidence->setEvalue(attributes.value("evalue").toDouble());
    _p_peptide_evidence->setExperimentalMass(attributes.value("exp_mass").toDouble());
    _p_peptide_evidence->setPeptideXtpSp(_map_peptides.at( attributes.value("peptide_id").simplified()));
    
    PeptideEvidenceSp sp_peptide_evidence = sp_ident_source.get()->getPeptideEvidenceStore().getInstance(_p_peptide_evidence);
    _map_peptide_evidences.insert(std::pair<QString, PeptideEvidenceSp>(attributes.value("id").simplified(), sp_peptide_evidence));
    qDebug() << "startElement_peptide_evidence end" ;
    return true;
}

bool XtpXpipSaxHandler::startElement_peptide_match(QXmlAttributes attributes) {

    qDebug() << "startElement_peptide_match ";
    /*
     * <protein_match acc="GRMZM2G138258_P01" checked="true">
                <peptide_match source_id="identa0" peptide_id="pd3533" scan="1907" rt="646.406" evalue="0.00044" exp_mass="2190.816924" start="103" charge="3" checked="true">
                    <param key="0" value="23.6"/>
                </peptide_match>
              */
    _p_protein_match = new ProteinMatch();
    qDebug() << "startElement_peptide_match end" ;
    return true;
}

bool XtpXpipSaxHandler::startElement_protein_match(QXmlAttributes attributes) {

    qDebug() << "startElement_protein_match ";
    /*
     * <protein_match acc="GRMZM2G138258_P01" checked="true">
                <peptide_match source_id="identa0" peptide_id="pd3533" scan="1907" rt="646.406" evalue="0.00044" exp_mass="2190.816924" start="103" charge="3" checked="true">
                    <param key="0" value="23.6"/>
                </peptide_match>
              */
    _p_protein_match = new ProteinMatch();
    _p_protein_match->setChecked(false);
    if (attributes.value("checked").simplified().toLower() == "true") {
        _p_protein_match->setChecked(true);
    }
    qDebug() << "startElement_protein_match end" ;
    return true;
}

bool XtpXpipSaxHandler::startElement_fasta_file(QXmlAttributes attributes) {
    //<fasta_file id="fastaa0" path="/espace/Genome_Z_mays_5a.fasta"/>
    qDebug() << "startElement_fasta_file ";

    FastaFileSp fasta_file_sp = _p_project->getFastaFileStore().getInstance(FastaFile(attributes.value("path")));
    fasta_file_sp.get()->setXmlId(attributes.value("id"));

    _map_fasta_files.insert(std::pair<QString, FastaFileSp>(attributes.value("id"), fasta_file_sp));
    qDebug() << "startElement_fasta_file end" ;
    return true;
}

bool XtpXpipSaxHandler::startElement_protein(QXmlAttributes attributes) {

    qDebug() << "startElement_protein ";
    /*
     *  <protein fasta_id="fastaa0" acc="AC149475.2_FGP003" description="seq=translation; coord=9:152490486..152495067:-1; parent_transcript=AC149475.2_FGT003; parent_gene=AC149475.2_FG003" is_decoy="false" is_contaminant="false">
            <sequence>MFLTRTEYDRGVNTFSPEGRLFQVEYAIEAIKLGSTAIGLKTKDGVVLAVEKRVTSPLLEPSSVEKIMEIDEHIGCAMSGLIADARTLVEHARVETQNHRFSYGEPMTVESSTQAICDLALRFGEGDEESMSRPFGVSLLIAGHDENGPSLYYTDPSGTFWQCNAKAIGSGSEGADSSLQEQYNKELALEEAETIALSILKQVMEEKVTPNNVDIAKVAPKYHLYTPAEVEAVIARL</sequence>
        </protein>
              */

    _current_protein.setFastaFileP(_map_fasta_files.at(attributes.value("fasta_id")).get());
    _current_protein.setAccession(attributes.value("acc"));
    _current_protein.setDescription(attributes.value("description"));
    _current_protein.setIsContaminant(false);
    if (attributes.value("is_contaminant").simplified().toLower() == "true") {
        _current_protein.setIsContaminant(true);
    }
    _current_protein.setIsDecoy(false);
    if (attributes.value("is_decoy").simplified().toLower() == "true") {
        _current_protein.setIsDecoy(true);
    }
    qDebug() << "startElement_protein end" ;
    return true;
}
bool XtpXpipSaxHandler::startElement_peptide(QXmlAttributes attributes) {

//    <peptide id="pd1493" seq="FEGVPMAPPDPILGVSEAFK">
//            <mod ref="moda1" position="6" aa="M"/>
//        </peptide>
    qDebug() << "startElement_peptide ";
    _current_peptide_sp = PeptideXtp(attributes.value("seq").simplified()).makePeptideXtpSp();

    _map_peptides.insert(std::pair<QString, PeptideXtpSp>(attributes.value("id").simplified(), _current_peptide_sp));
    qDebug() << "startElement_peptide end" ;
    return true;
}

bool XtpXpipSaxHandler::startElement_mod(QXmlAttributes attributes) {

    //<mod ref="moda1" position="6" aa="M"/>
    qDebug() << "startElement_mod ";
    pappso::AaModificationP modif = _map_modifs[attributes.value("ref").simplified()];
    unsigned int position = attributes.value("position").simplified().toUInt();
    _current_peptide_sp.get()->addAaModification(modif, position);
    qDebug() << "startElement_mod end" ;
    return true;
}
bool XtpXpipSaxHandler::endElement_peptide() {
    qDebug() << "endElement_peptide ";

    return true;
}

bool XtpXpipSaxHandler::endElement_sequence() {
    qDebug() << "endElement_sequence begin ";
    //if ((_tag_stack.size() > 1) && (_tag_stack[_tag_stack.size() - 1] == "protein")) {

    //._sequence.replace(QRegExp("\\*"), "")).removeTranslationStop()
    _current_protein.setSequence(_current_text);
    //}
    //else {
    // XtandemHyperscore hyperscore(_curent_spectrum, _current_peptide_sp, _precision, _ion_list, _max_charge,_refine_spectrum_synthesis);
    //}
    qDebug() << "endElement_sequence end ";
    return true;
}
bool XtpXpipSaxHandler::endElement_protein() {

    ProteinXtpSp sp_xtp_protein = _current_protein.makeProteinXtpSp();
    _map_proteins.insert(std::pair<QString, ProteinXtpSp>(sp_xtp_protein.get()->getAccession(), sp_xtp_protein));
    return true;
}

bool XtpXpipSaxHandler::endElement_protein_match() {
    _count_protein_match++;
    _p_monitor->message(QObject::tr("reading match %1").arg(_count_protein_match), _count_protein_match);
    _current_identification_group_p->addProteinMatch(_p_protein_match);
    _p_protein_match = nullptr;
    return true;
}


bool XtpXpipSaxHandler::error(const QXmlParseException &exception) {
    _errorStr = QObject::tr("Parse error at line %1, column %2 :\n"
                            "%3").arg(exception.lineNumber()).arg(exception.columnNumber()).arg(
                    exception.message());

    return false;
}


bool XtpXpipSaxHandler::fatalError(const QXmlParseException &exception) {
    _errorStr = QObject::tr("Parse error at line %1, column %2 :\n"
                            "%3").arg(exception.lineNumber()).arg(exception.columnNumber()).arg(
                    exception.message());
    return false;
}

QString XtpXpipSaxHandler::errorString() const {
    return _errorStr;
}


bool XtpXpipSaxHandler::endDocument() {
    //_p_project->getProteinStore().setRegexpDecoyPattern(_p_project->getProteinStore().getRegexpDecoy().pattern());
    _p_project->updateAutomaticFilters(_automatic_filter_parameters);
    return true;
}

bool XtpXpipSaxHandler::startDocument() {
    _count_protein_match=0;
    return true;
}

bool XtpXpipSaxHandler::characters(const QString &str) {
    _current_text += str;
    return true;
}

